/** 
 * @brief "User interface to configure the HCTL chip on wizhctl boards"
 * 
 * @file config.c
 * @author ch'gans <chgans@tuxfamily.org>
 * @date 30 sept 2003
 * @version $Id: config.c,v 1.1 2003/10/06 00:23:43 chgans Exp $ 
 * \par License:
 * Copyright (C) WISIRC <http://wisirc.tuxfamily.org>\n
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <argp.h>
#include <string.h>

#include <sys/ioctl.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <wizhctl.h>

#include "common.h"

const char *program_name =	"wizhctl-config";
const char *program_version =	"wizhctl-config version " VERSION;

static char doc[] =
"\
Options for configuring sample timer:\n\
\t[-vqs] -T [-d DEVICE] -f FREQ -t PERIOD\n\
Options for configuring filter:\n\
\t[-vqs] -F [-d DEVICE] -z ZERO -p POLE -k GAIN\
\v\
\"wizhctl-config\" from wizhctl-utils package, by WISIRC's WizMotion team:\n\
\tA user interface to configure the HCTL chip on wizhctl boards\n\
(c) 2003 WISIRC <http://wisirc.tuxfamily.org>\
";


typedef enum
{
  CONFIG_NONE = 0,
  CONFIG_TIMER = 1,
  CONFIG_FILTER = 2
} config_t;

static struct argp_option options[] = {
  /* device for use */
  {"device",	'd',	"DEVICE",	0,	"Use DEVICE to access HCTL board (default = " DEFAULT_DEVICE ")\n", 1 },
  /* Config timer */
  {"timer",	'T',	0,		0,	"Configure HCTL chip sample timer (need -f and -t)", 1 },
  {"clkfreq",	'f',	"FREQ",		0,	"Assume HCTL chip is clocked at FREQ MHz (default = 2MHz).", 2 },
  {"smptime",	't',	"PERIOD",	0,	"Set the HCTL chip sample timer period to (PERIOD*FREQ/16 -1) us.\n", 2 },
  /* Config filter */
  {"filter",	'F',	0,		0,	"Configure HCTL chip filter (need -z, -p and -k)", 3 },
  {"fzero",	'z',	"ZERO",		0,	"Set the HCTL chip filter zero to ZERO/256.\n", 4 },
  {"fpole",	'p',	"POLE",		0,	"Set the HCTL chip filter pole to POLE/256.", 4 },
  {"fgain",	'k',	"GAIN",		0,	"Set the HCTL chip filter gain to GAIN/4.", 4 },
  /* verbosity */
  {"verbose",	'v',	0,		0,	"Produce verbose output", 6 },
  {"quiet",	'q',	0,		0,	"Don't produce any output", 6 },
  {"silent",	's',	0,		OPTION_ALIAS },
  { 0 }
};

/* Communication between main and parse_opt */
struct arguments
{
  int		verbose;
  config_t	cfgtype;
  u_int16_t	smptime;
  int		smptime_passed;
  float		clkfreq;
  int		clkfreq_passed;
  u_int8_t	fzero;
  int		fzero_passed;
  u_int8_t	fpole;
  int		fpole_passed;
  u_int8_t	fgain;
  int		fgain_passed;
  char		*device;
};

static error_t parse_opt (int key, char *arg, struct argp_state *state)
{
  /* Get the input argument from argp_parse, which we
     know is a pointer to our arguments structure. */
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'q': case 's':
      arguments->verbose = 0;
      break;

    case 'v':
      arguments->verbose = 1;
      break;

    case 'd':
      printf("%s\n", arg);
      arguments->device = arg;
      break;

    case 'T':
      if(arguments->cfgtype != CONFIG_NONE)
        argp_usage (state);
      else
	arguments->cfgtype = CONFIG_TIMER;
      break;

    case 'f':
      /* TODO: autoriser une frequence avec une virgule !!! */
      sscanf(arg, "%f", &arguments->clkfreq);
      arguments->clkfreq_passed = 1;
      break;

    case 't':
      sscanf(arg, "%hd", &arguments->smptime);
      arguments->smptime_passed = 1;
      break;

    case 'F':
      if(arguments->cfgtype != CONFIG_NONE)
        argp_usage (state);
      else
	arguments->cfgtype = CONFIG_FILTER;
      break;

    case 'z':
      sscanf(arg, "%hd", &arguments->fzero);
      arguments->fzero_passed = 1;
      break;

    case 'k':
      sscanf(arg, "%hd", &arguments->fgain);
      arguments->fgain_passed = 1;
      break;

    case 'p':
      sscanf(arg, "%hd", &arguments->fpole);
      arguments->fpole_passed = 1;
      break;

    case ARGP_KEY_ARG:
      if (state->arg_num != 0)
        /* No arguments needed! */
        argp_usage (state);
      break;

    case ARGP_KEY_END:
#if 0
      printf("type=%d\ns=%d, f=%d\nz=%d, p=%d, k=%d\n",
	     arguments->cfgtype,
	     arguments->smptime,
	     arguments->clkfreq,
	     arguments->fzero,
	     arguments->fpole,
	     arguments->fgain);
#endif
      switch(arguments->cfgtype)
	{
	case CONFIG_TIMER:
	  if( (arguments->smptime_passed != 1) ||
	      ( (arguments->fpole_passed != 0) &&
		(arguments->fzero_passed != 0) &&
		(arguments->fgain_passed != 0) ) )
	    argp_usage(state);
	  break;

	case CONFIG_FILTER:
	  if( ( (arguments->fzero_passed != 1) ||
		(arguments->fpole_passed != 1) ||
		(arguments->fgain_passed != 1) ) ||
	      ( (arguments->smptime_passed != 0) &&
		(arguments->clkfreq_passed != 0) ) )
	    argp_usage(state);
	  break;

	case CONFIG_NONE:
	default:
	  argp_usage (state);
	  break;
	}
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = { options, parse_opt, NULL, doc, NULL, NULL, NULL };



int main(int argc, char **argv)
{
  struct arguments args;
  int fd;
  int ioctlnb;
  struct hctl_filter filter;
  void *pdata;
  float timeval;

  /* Valeurs par dfault */
  args.verbose = 1;
  args.cfgtype = CONFIG_NONE;
  args.smptime = 0;
  args.smptime_passed = 0;
  args.clkfreq = DEFAULT_CLOCK_FREQ;
  args.clkfreq_passed = 0;
  args.fzero = 0;
  args.fzero_passed = 0;
  args.fpole = 0;
  args.fpole_passed = 0;
  args.fgain = 0;
  args.fgain_passed = 0;
  args.device = DEFAULT_DEVICE;

  /* Parse our arguments; every option seen by parse_opt will
     be reflected in arguments. */
  argp_program_bug_address = bug_address;
  argp_program_version = program_version;
  argp_parse (&argp, argc, argv, 0, 0, &args);

  
  fd = open(args.device,  O_RDWR | O_SYNC);
  if (fd <= 0) {
    perror(args.device);
    exit(-1);
  }

  switch(args.cfgtype)
    {
    case CONFIG_TIMER:
      ioctlnb = HCTL_IOC_INIT_TIMER;
      timeval = 16*(args.smptime + 1)/args.clkfreq;
      pdata = (void*)&timeval;
      if(args.verbose == 1)
	{
	  printf("Sample timer period with T=%d and F=%.3f:\n",
		 args.smptime,
		 args.clkfreq);
	  printf("\t%.3f us\n", 
		 timeval);
	  
		 
	}
      break;

    case CONFIG_FILTER:
      ioctlnb = HCTL_IOC_INIT_FILTER;
      filter.zero = args.fzero;
      filter.pole = args.fpole;
      filter.gain = args.fgain;
      pdata = (void*)&filter;
      if(args.verbose == 1)
	{
	  printf("Filter with K=%d, Z=%d and P=%d:\n",
		 filter.gain, 
		 filter.zero, 
		 filter.pole);
	  printf("        ( z - %1.3f )\n", (float)((float)filter.zero/256));
	  printf("%3.3f . -------------\n", (float)((float)filter.gain/4));
	  printf("        ( z + %1.3f )\n", (float)((float)filter.pole/256));
	  printf("\n");
	  /*
	  printf("Sample timer:\n\tvalue: %d us\n\tregister value: 0x%02X (%d)\n", 
		 args.smptime,
		 timeval,
		 timeval);
	  */
	}
      break;
    case CONFIG_NONE:
    default:
      ERROR("Config type???\n");
      close(fd);
      exit(-1);      
    }

  if (ioctl(fd, ioctlnb, pdata) < 0)
    {
      perror("IOCTL failed!");
      close(fd);
      exit(-1);
    }

  /* That's all folks! */
  close(fd);
  return 0;
}





















#if 0
static char args_doc[] = "";

static char *help_texte;

static char *help_filter(int key, const char *text, void *input)
{
  return text;
  help_texte = malloc(256);
  switch(key)
    {
    case ARGP_KEY_HELP_PRE_DOC:      
      //strcpy(help_texte,"ARGP_KEY_HELP_PRE_DOC");
      return text;
      break;
    case ARGP_KEY_HELP_POST_DOC:
      strcpy(help_texte,"ARGP_KEY_HELP_POST_DOC");
      break;
    case ARGP_KEY_HELP_HEADER:
      strcpy(help_texte,"ARGP_KEY_HELP_HEADER");
      break;
    case ARGP_KEY_HELP_EXTRA:
      strcpy(help_texte,"ARGP_KEY_HELP_EXTRA");
      break;
    case ARGP_KEY_HELP_DUP_ARGS_NOTE:
      strcpy(help_texte,"ARGP_KEY_HELP_DUP_ARGS_NOTE");
      break;
    case ARGP_KEY_HELP_ARGS_DOC: 
      strcpy(help_texte,"ARGP_KEY_HELP_ARGS_DOC");
      break;
    default:
      strcpy(help_texte,"default");
      break;
    }
  return help_texte;
}


#if 0
  int filter_done;
  int filter_begin;
  int timer_done;
  int timer_begin;
  int none_done;
  int some_begin;

      timer_begin = arguments->smptime_passed || arguments->clkfreq_passed;
      timer_done = arguments->smptime_passed && arguments->clkfreq_passed;
      timer_ok = timer_begin && timer_done;

      filter_begin = arguments->fzero_passed || arguments->fpole_passed || arguments->fgain_passed;
      filter_done = arguments->fzero_passed && arguments->fpole_passed && arguments->fgain_passed;
      filter_ok = filter_begin && filter_done; 

      some_begin = timer_begin || filter_begin;
      none_done = !timer_done && !filter_done;

      /* */
      if( ( (filter_begin == 1) && 
	    (timer_begin == 1 ) ) ||
	  ( (filter_done == 0) && 
	    (timer_done == 0 ) ) ) {
	argp_usage (state);
      }
      
      if( (filter_begin == 1) && (filter_done == 0) ) {
	printf("%d\n", __LINE__);
	argp_usage (state);
      }

      if( (timer_begin == 1) && (timer_done == 0) ) {
	printf("%d\n", __LINE__);
	argp_usage (state);
      }
#endif

#if 0
      if ( (arguments->smptime_passed == 1) || 
	   (arguments->clkfreq_passed == 1) ) {
	timer_begin = 1;
	printf("%d\n", __LINE__);
      }
      else {
	timer_begin = 0;
	printf("%d\n", __LINE__);
      }

      if ( (arguments->smptime_passed == 1) && 
	   (arguments->clkfreq_passed == 1)) {
	timer_done = 1;
	printf("%d\n", __LINE__);
      }
      else {
	timer_done = 0;
	printf("%d\n", __LINE__);
      }

      if ( (arguments->fzero_passed == 1) || 
	   (arguments->fpole_passed == 1) || 
	   (arguments->fgain_passed == 1)) {
	filter_begin = 1;
	printf("%d\n", __LINE__);
      }
      else  {
	filter_begin = 0;
	printf("%d\n", __LINE__);
      }

      if ( (arguments->fzero_passed == 1) && 
	   (arguments->fpole_passed == 1) && 
	   (arguments->fgain_passed == 1)) {
	filter_done = 1;
	printf("%d\n", __LINE__);
      }
      else {
	filter_done = 0;
	printf("%d\n", __LINE__);
      }

      if( ( (filter_begin == 1) && 
	    (timer_begin == 1 ) ) )
	printf("%d\n", __LINE__);
	
      if ( (filter_done == 0) && 
	    (timer_done == 0 ) )
	printf("%d\n", __LINE__);

      if( ( (filter_begin == 1) && 
	    (timer_begin == 1 ) ) ||
	  ( (filter_done == 0) && 
	    (timer_done == 0 ) ) ) {
	printf("%d\n", __LINE__);
	argp_usage (state);
      }
#endif

#if 0
      if( ( (filter_begin == 1) && 
	    (timer_begin == 1 ) ) ||
	  ( (filter_done == 0) && 
	    (timer_done == 0 ) ) ) {
	argp_usage (state);
      }
      
      if( (filter_begin == 1) && (filter_done == 0) ) {
	printf("%d\n", __LINE__);
	argp_usage (state);
      }

      if( (timer_begin == 1) && (timer_done == 0) ) {
	printf("%d\n", __LINE__);
	argp_usage (state);
      }
#endif

#endif
