/** 
 * @brief "User interface to make motor(s) measures on wizhctl boards"
 * 
 * @file motorcmd.c
 * @author ch'gans <chgans@tuxfamily.org>
 * @date 30 sept 2003
 * @version $Id$ 
 * \par License:
 * Copyright (C) 2003 WISIRC <http://wisirc.tuxfamily.org>\n
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <argp.h>

#include <sys/ioctl.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <wizhctl.h>

#include "common.h"

 
const char *program_name =	"wizhctl-measures";
const char *program_version =	"wizhctl-measures version "VERSION;

static char doc[] =
"\
wizhctl-measures [-sqv] [-d DEVICE] -p|-l|-u CONS -t PERIOD -n N [-o FILE]\n\
\v\
Note:\n\
\tWhen using PWM command port, HCTL sample timer period has effect on PWM frequency, \
to fix this parameter you can use wizhctl-config (try wizhctl-config --help for more information)\n\
\n\
\"wizhctl-measures\" from wizhctl-utils package, by WISIRC's WizMotion team:\n\
\tA user interface to make motor(s) measures on wizhctl boards.\n\
(c) 2003 WISIRC <http://wisirc.tuxfamily.org>\n\
";

//static char args_doc[] = "";

static struct argp_option options[] = {
  {"consigne",	'u',	"+/-CONS",	0,	"Set the consigne on PWM motor command port to +/-CONS% (!=0)", 1 },
  {"pwmport",	'p',	"+/-CONS",	0,	"Apply consigne CONS to the PWM command port (not yet implemented, use -u)", 1 },
  {"parport",	'l',	"+/-CONS",	0,	"Apply consigne CONS to the 8 bits command port (not yet implemented, use -u)", 1 },
  {"period",	't',	"PERIOD",	0,	"Set period of samples to PERIOD ms (>0)", 2 },
  {"nbsample",	'n',	"N",		0,	"Set the number of samples to N (>0)", 2 },
  {"device",	'd',	"DEVICE",	0,	"Use DEVICE to access HCTL board (default = " DEFAULT_DEVICE ")", 3 },
  {"output",	'o',	"FILE",		0,	"Output to FILE instead of standard output", 3 },
  {"verbose",	'v',	0,		0,	"Produce verbose output", 6 },
  {"quiet",	'q',	0,		0,	"Don't produce any output", 6 },
  {"silent",	's',	0,		OPTION_ALIAS },
  { 0 }
};

struct arguments
{
  int		verbose;
  u_int32_t	period;
  int8_t	consigne;
  u_int32_t	nbsample;
  char		*output_file;
  char		*device;
};

static error_t parse_opt (int key, char *arg, struct argp_state *state)
{
  /* Get the input argument from argp_parse, which we
     know is a pointer to our arguments structure. */
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'q': case 's':
      arguments->verbose = 0;
      break;

    case 'v':
      arguments->verbose = 1;
      break;

    case 'o':
      arguments->output_file = arg;
      break;

    case 't':
      sscanf(arg, "%lu", &arguments->period);
      if (!(arguments->period > 0)) {
	ERROR("period <= 0\n");
        argp_usage (state);	
      }
      break;

    case 'u':
      sscanf(arg, "%hd", &arguments->consigne);
      break;

    case 'n':
      sscanf(arg, "%hu", &arguments->nbsample);
      if (!(arguments->nbsample > 0)) {
	ERROR("nbsamples <= 0\n");
        argp_usage (state);	
      }
      break;

    case 'd':
      printf("%s\n", arg);
      arguments->device = arg;
      break;

    case ARGP_KEY_ARG:
      if (state->arg_num != 0)
        /* No arguments needed! */
        argp_usage (state);
      break;

    case ARGP_KEY_END:
  /* Vrification des options */
      if (arguments->period <= 0 ||
	  arguments->consigne <=0 ||
	  arguments->nbsample <= 0)
	argp_usage(state);
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = { options, parse_opt, NULL, doc, NULL, NULL, NULL };


int main(int argc, char **argv)
{
  struct arguments args;
  wizhctl_motorcmd_measure_t measure;
  FILE *stream;
  int fd;
  int i;

  /* Default values. */
  args.verbose = 0;
  args.output_file = NULL;
  args.period = 0;
  args.nbsample = 0;
  args.consigne = 0;
  args.device = DEFAULT_DEVICE;

  /* Parse our arguments; every option seen by parse_opt will
     be reflected in arguments. */
  argp_program_bug_address = bug_address;
  argp_program_version = program_version;
  argp_parse (&argp, argc, argv, 0, 0, &args);

  fd = open(args.device,  O_RDWR | O_SYNC);
  if (fd <= 0) {
    perror(args.device);
    exit(-1);
  }
  
  if (args.output_file == NULL) {
    stream = stdout;
  }
  else {
    stream = fopen(args.output_file, "w");
    if (stream == NULL) {
      perror(args.output_file);
      exit(-1);
    }
  }

  measure.consigne = args.consigne;
  measure.periode = args.period;
  measure.nb_measure = args.nbsample;
  measure.position = (int32_t*)malloc(measure.nb_measure*sizeof(*measure.position));
  measure.time = (struct timeval*)malloc(measure.nb_measure*sizeof(*measure.time));
  measure.velocity = (int16_t*)malloc(measure.nb_measure*sizeof(*measure.velocity));

  if (ioctl(fd, HCTL_IOC_MOTORCMD_MEASURE, &measure) < 0)
    {
      perror("Measure failed!");
      close(fd);
      free(measure.position);
      free(measure.time);
      exit(-1);
    }

#if 0
  /* Print velocity vector */
  for(i=0; i<measure.nb_measure; i++) {
    fprintf(stream, "%d ", measure.velocity[i]);
  }
  fprintf(stream, "\n");
#endif

  /* Print position vector */
  for(i=0; i<measure.nb_measure; i++) {
    fprintf(stream, "%d ", measure.position[i]);
  }
  fprintf(stream, "\n");

  /* Print time vector */
  for(i=0; i<measure.nb_measure; i++) {
    fprintf(stream, "%lu%lu ", measure.time[i].tv_sec, measure.time[i].tv_usec);
  }
  fprintf(stream, "\n");

  /* That's all folks! */
  close(fd);
  fclose(stream);
  free(measure.position);
  free(measure.time);
  
  return 0;
}
