/** 
 * @brief "User interface to use wizhctl boards in proportional velocity mode"
 * 
 * @file pvelocity.c
 * @author ch'gans <chgans@tuxfamily.org>
 * @date 06 oct 2003
 * @version $Id: velocity.c,v 1.1 2003/10/06 00:23:43 chgans Exp $ 
 * \par License:
 * Copyright (C) WISIRC <http://wisirc.tuxfamily.org>\n
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <argp.h>
#include <termios.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <wizhctl.h>

#include "common.h"

int interactive_mode(int fd, float initial_velocity, int verbose);
int batch_mode(int fd, char* ifilename, char* ofilename, u_int32_t period, float initial_velocity, int verbose);
void interactive_mode_inline_help(void);
void batch_mode_file_help(void);
 
const char *program_name =	"wizhctl-pvelocity";
const char *program_version =	"wizhctl-pvelocity version "VERSION;

static char doc[] =
"\
Options for interactive mode:
\t[-sqv] [-d DEVICE] [-I IVELOCITY]
Options for \"batch\" mode:
\t[-sqv] [-d DEVICE] [-I IVELOCITY] -t PERIOD -i IFILE [-o OFILE]
\v\
Notes:\n\
- Velocity precision is 1/16 (velocity is internally multiply by 16 before writing it in HCTL register).\n\
- Velocity unit is quadrature counts per sample time. To convert from rpm \
to quadrature counts per sample time, use the formula shown below:\n\
\tVq = (Vr).(N).(t).1/60\n\
\tWhere :\n\
\t\tVq = velocity in quadrature counts per sample time.\n\
\t\tVr = velocity in RPM.\n\
\t\tN  = 4 times the number of slots in the codewheel (ie., quadrature counts).\n\
\t\tt  = The HCTL100 sample time in seconds.\n\
- In proportional velocity mode, only the gain factor K is used  to perform the control of the motor speed.\n\
- You can configure the HCTL1100 sample timer and filter with wizhctl-config (try wizhctl-config --help for more information).\n\
- Formats of input file:\n\
\tFile must begin with a comment line containing the number of commands and optionally (when using -o OFILE) the sample period of measures:\n\
\t\t//<NB_COMMAND>[ <SAMPLE_PERIOD>]\n<CMD1> <CMD2> ....\n<DUR1> <DUR2> ...\n\
- Formats of onput file:\n\
\tWhen using -o OFILE, measures are enable, the format of the output file is:\n\
\t\t//<NB_MEASURES>\n<MES1> <MES2> ...\n\
\n\
\"wizhctl-pvelocity\" from wizhctl-utils package, by WISIRC's WizMotion team:\n\
\tA user interface to use wizhctl boards in proportional velocity mode.\n\
(c) 2003 WISIRC <http://wisirc.tuxfamily.org>\n\
";

static struct argp_option options[] = {
  {"velocity",	'I',	"IVELOCITY",	0,	"Set initial velocity to IVELOCITY, (default is 0.0)", 3},
  {"input",	'i',	"IFILE",	0,	"Read velocity commands from IFILE (also know as batch mode)", 4},
  {"output",	'o',	"OFILE",	0,	"Save actual velocity and measure time points in OFILE (works only in batch mode)", 4 },
  {"device",	'd',	"DEVICE",	0,	"Use DEVICE to access HCTL board (default = " DEFAULT_DEVICE ")", 5 },
  {"verbose",	'v',	0,		0,	"Produce verbose output", 6 },
  {"quiet",	'q',	0,		0,	"Don't produce any output", 6 },
  {"silent",	's',	0,		OPTION_ALIAS },
  { 0 }
};


struct arguments
{
  int		verbose;
  float		initial_velocity;
  u_int32_t	period;
  char		*output_file;
  char		*input_file;
  char		*device;
};

static int16_t pvelocity2reg(float velocity)
{
  if( ((16*velocity) > (float)32767) || ((16*velocity) < (float)(-32768)) ) {
    WARNING("%f: Velocity out of range (2047.9 < Vq < -2048)\n", velocity);
  }

  return (int16_t)(16*velocity);
}

/** Parser d'options */
static error_t parse_opt (int key, char *arg, struct argp_state *state)
{
  /* Get the input argument from argp_parse, which we
     know is a pointer to our arguments structure. */
  struct arguments *arguments = state->input;
  float velocity;

  switch (key)
    {
    case 'q': case 's':
      arguments->verbose = 0;
      break;

    case 'v':
      arguments->verbose = 1;
      break;

    case 'o':
      arguments->output_file = arg;
      break;

    case 'i':
      arguments->input_file = arg;
      break;

    case 'I':
      sscanf(arg, "%f", &velocity);
      arguments->initial_velocity = velocity;
      break;

    case 'd':
      printf("%s\n", arg);
      arguments->device = arg;
      break;

    case ARGP_KEY_ARG:
      if (state->arg_num != 0)
        /* No arguments needed! */
        argp_usage (state);
      break;

    case ARGP_KEY_END:
      /* Vrification des options */
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = { options, parse_opt, NULL, doc, NULL, NULL, NULL };


/*
 * Gestion du terminal
 */

struct termios saved_attributes;

static void term_reset_input_mode (void)
{
  tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
}

static void term_set_input_mode (void)
{
  struct termios tattr;
  //char *name;

  /* Make sure stdin is a terminal. */
  if (!isatty (STDIN_FILENO))
    {
      fprintf (stderr, "Not a terminal.\n");
      exit (EXIT_FAILURE);
    }

  /* Save the terminal attributes so we can restore them later. */
  tcgetattr (STDIN_FILENO, &saved_attributes);
  atexit (term_reset_input_mode);

  /* Set the funny terminal modes. */
  tcgetattr (STDIN_FILENO, &tattr);
  tattr.c_lflag &= ~(ICANON|ECHO); /* Clear ICANON and ECHO. */
  tattr.c_cc[VMIN] = 1;
  tattr.c_cc[VTIME] = 0;
  tcsetattr (STDIN_FILENO, TCSAFLUSH, &tattr);
}

/*
 * Mode intractif
 */
void interactive_mode_inline_help(void)
{
  printf("[interactive mode]\n");
  printf("Usage:\n");
  printf("\t'i' to increment velocity by 1.\n");
  printf("\t'I' to increment velocity by 10.\n");
  printf("\t'd' to decrement velocity by 1.\n");
  printf("\t'D' to decrement velocity by 10.\n");
  printf("\t'z' to reset velocity to 0.\n");
  printf("\t'r' to print command and actual velocity.\n");
  printf("\t'h' to print this help.\n");
  printf("\t'q' to quit.\n");
  printf("\t'v' to turn ON/OFF verbosity\n");
}

int interactive_mode(int fd, float initial_velocity, int verbose)
{
  char key;
  float command_velocity;
  int16_t command_velocity_reg;
  int16_t actual_velocity;
  int32_t position;
  int lverbose=verbose;

  /* initialize terminal */
  term_set_input_mode();

  /* print short help */
  if(lverbose)
    {
      interactive_mode_inline_help();
    }

  /* Command position  0 */
  position = 0;
  if (ioctl(fd, HCTL_IOC_SET_COMMAND_POSITION, &position) < 0)
    {
      perror("HCTL_IOC_SET_COMMAND_POSITION");
      return -1;
    }
  /* Actual position  0 */
  if (ioctl(fd, HCTL_IOC_RESET_ACTUAL_POSITION, NULL) < 0)
    {
      perror("HCTL_IOC_RESET_ACTUAL_POSITION");
      return -1;
    }
  /* Go to position mode */
  if (ioctl(fd, HCTL_IOC_POSITION_MODE, NULL) < 0)
    {
      perror("HCTL_IOC_POSITION_MODE");
      return -1;
    }
  /* Velocity = init */
  command_velocity = actual_velocity = initial_velocity;
  command_velocity_reg = pvelocity2reg((float)command_velocity);
  if (ioctl(fd, HCTL_IOC_SET_PROPORTIONAL_VELOCITY, &command_velocity_reg) < 0)
    {
      perror("HCTL_IOC_SET_PROPORTIONAL_VELOCITY");
      return -1;
    }
  /* Go to proportional velocity mode */
  if (ioctl(fd, HCTL_IOC_START_PROPORTIONAL_VELOCITY, NULL) < 0)
    {
      perror("HCTL_IOC_START_PROPORTIONAL_VELOCITY");
      return -1;
    }

  /* Test en boucle clavier / registres " la main" */
  read (STDIN_FILENO, &key, 1);
  while(key!='q')
    {
      switch(key)
	{
	case 'z':
	  command_velocity = 0;
	  break;
	case 'i':
	  command_velocity++;
	  break;
	case 'd':
	  command_velocity--;
	  break;
	case 'I':
	  command_velocity+=10;
	  break;
	case 'D':
	  command_velocity-=10;
	  break;
	case 'r': /* refresh */
	  break;
	case 'v':
	  if (lverbose == 0) 
	    {
	      lverbose = 1 ;
	      printf("Verbosity is ON.\n");
	    }
	  else
	    {
	      lverbose = 0;
	      printf("Verbosity is OFF.\n");
	    }
	  break;
	case 'h':
	  interactive_mode_inline_help();
	  break;
	default:
	  printf("Unknow key '%c' (0x%02X)\n", key, key);
	  break;
	}
      switch(key)
	{
	case 'z':
	case 'i':
	case 'd':
	case 'I':
	case 'D':
	  command_velocity_reg = pvelocity2reg((float)command_velocity);
	  if (ioctl(fd, HCTL_IOC_SET_PROPORTIONAL_VELOCITY, &command_velocity_reg) < 0)
	    {
	      perror("HCTL_IOC_SET_PROPORTIONAL_VELOCITY");
	      return -1;
	    }
	  if(lverbose)
	    {
	      if (ioctl(fd, HCTL_IOC_READ_PROPORTIONAL_VELOCITY, &actual_velocity) < 0)
		{
		  perror("HCTL_IOC_READ_PROPORTIONAL_VELOCITY");
		  return -1;
		}
	      printf("%c => Command Velocity = % 5.4f\tActual velocity = %5d\n", 
		     key, command_velocity, actual_velocity);
	    }
	  break;
	case 'r':
	  if (ioctl(fd, HCTL_IOC_READ_PROPORTIONAL_VELOCITY, &actual_velocity) < 0)
	    {
	      perror("HCTL_IOC_READ_PROPORTIONAL_VELOCITY");
	      return -1;
	    }
	  printf("%c => Command Velocity = % 5.4f\tActual velocity = %5d\n", 
		 key, command_velocity, actual_velocity);
	  break;
	}
      read(STDIN_FILENO, &key, 1);
    }

  /* Stop proportional velocity mode */
  command_velocity = 0;
  actual_velocity = 0;
  if (ioctl(fd, HCTL_IOC_STOP_PROPORTIONAL_VELOCITY, NULL) < 0)
    {
      perror("HCTL_IOC_STOP_PROPORTIONAL_VELOCITY");
      return -1;
    }
  /* Go to idle mode */
  if (ioctl(fd, HCTL_IOC_IDLE_MODE, NULL) < 0)
    {
      perror("HCTL_IOC_POSITION_MODE");
      return -1;
    }
  return 0;
}

void batch_mode_file_help(void)
{
  printf("File must begin with a comment line containing the number of commands and optionally the sample period of measures:\n");
  printf("//<NB_COMMAND>[ <SAMPLE_PERIOD>]\n<CMD1> <CMD2> ....\n<DUR1> <DUR2> ...\n\n");
  return;
}

/*
 * Batch mode.
 */
int batch_mode(int fd, char* ifilename, char* ofilename, u_int32_t period, float initial_velocity, int verbose)
{
  int32_t position;
  int lverbose=verbose;
  float command_velocity;
  int16_t command_velocity_reg;
  //int16_t actual_velocity;
  wizhctl_pvelocity_measure_t measure;
  u_int32_t total_time;
  FILE* istream;
  FILE* ostream;
  //struct stat filestat;
  int index;
  int error;
  float dummyf;
  //int nb_cmd;

  /* ouverture du fichier de commande */
  istream = fopen(ifilename, "r");
  if (istream == NULL) {
    perror(ifilename);
    return -1;
  }

  /* on recherche le nombre de commandes et ventuellement 
   * la priode d'chantillonage des mesures */
  error = fscanf(istream, "//%d %d\n", &measure.nb_command, &measure.measure_period);
  if( error == 1 )
    {
      if(ofilename != NULL)
	{
	  ERROR("\"-o OFILE\" was used but no measure sample period was found in %s.\n", ifilename);
	  batch_mode_file_help();
	  fclose(istream);
	  return -1;
	}
      measure.measure_period = 0;
    } 
  else if (error == 2 )
    {
      if(ofilename == NULL)
	{
	  if(measure.measure_period != 0) 
	    {
	      if(lverbose>0)
		WARNING("I've found a measure sample period in %s but no \"-o OFILE\" was used. Measure is OFF.\n", ifilename);
	      measure.measure_period = 0;
	    }
	}
    }
  else
    {
      ERROR("Unexpected error while reading %s for nb commands and measure sample period\n", ifilename);
      batch_mode_file_help();
      fclose(istream);
      return -1;
    }
  
  /* On alloue les buffers pour les commandes */
  measure.command_velocity = (int16_t*)malloc(sizeof(*measure.command_velocity)*measure.nb_command);
  if(measure.command_velocity==NULL)
    {
      ERROR("");
      perror("");
      fclose(istream);
      return -1;
    }

  measure.command_duration = (u_int32_t*)malloc(sizeof(*measure.command_duration)*measure.nb_command);
  if(measure.command_duration==NULL)
    {
      ERROR("");
      perror("");
      free(measure.command_velocity);
      fclose(istream);
      return -1;
    }

  /* On rcupre les commandes dans notre buffer ainsi que les duration et 
     en cumule les temps pour savoir le nombre de mesures */
  index = 0;
  for( index = 0; index < measure.nb_command ; index++)
    {
      error = fscanf(istream, "%f", &command_velocity);
      //printf("%f\n", command_velocity);
      if(error != 1)
	{
	  WARNING("Malformed input file: read %f\n", command_velocity);
	  ERROR("Unexpected error while reading %s for velocity\n", ifilename);
	  fclose(istream);
	  free(measure.command_velocity);
	  free(measure.command_duration);
	  return -1;
	}
      measure.command_velocity[index] = pvelocity2reg(command_velocity);
    }

  /* FIXME tester index /nb_command */
  //printf("index after command: %d\n", index);

  total_time = 0;
  index = 0;
  for( index = 0; index < measure.nb_command ; index++)
    {
      error = fscanf(istream, "%f", &dummyf);
      if(error != 1)
	{
	  WARNING("Malformed input file: read %d\n", measure.command_duration[index]);
	  ERROR("Unexpected error while reading %s for durations\n", ifilename);
	  fclose(istream);
	  free(measure.command_velocity);
	  free(measure.command_duration);
	  return -1;
	}
      measure.command_duration[index]=(u_int32_t)dummyf;
      //printf("command duration[%d] = %d\n", index, measure.command_duration[index]);
      total_time += measure.command_duration[index];
    }

  /* FIXME tester index /nb_command */
  
  /* On alloue  le buffer des vitesse mesures */
  if( measure.measure_period > 0 )
    {
      /* commandes en ms, mesures en s */ 
      measure.nb_measure = total_time*1000/measure.measure_period;
      measure.actual_velocity = (int16_t*)malloc(sizeof(*measure.actual_velocity)*measure.nb_measure);
      if(measure.actual_velocity==NULL)
	{
	  perror("");
	  free(measure.command_velocity);
	  free(measure.command_duration);
	  fclose(istream);
	  return -1;
	}
      measure.position = (int32_t*)malloc(sizeof(*measure.position)*measure.nb_measure);
      if(measure.position==NULL)
	{
	  perror("");
	  free(measure.command_velocity);
	  free(measure.command_duration);
	  free(measure.actual_velocity);
	  fclose(istream);
	  return -1;
	}
    }
  else
    {
      measure.nb_measure = 0;
    }

  if(lverbose>0)
    {
      printf("[batch mode] Read %d commands from %s (%d measures with period = %d S)\n", 
	     measure.nb_command, ifilename, measure.nb_measure, measure.measure_period);
      fflush(NULL);
    }

  /* on en a finit avec le fichier de commande */
  fclose(istream);

  //return -1;

  /* 
   * Y'a plus qu' ! 
   */

  error =-1;

  /* Command position  0 */
  position = 0;
  if (ioctl(fd, HCTL_IOC_SET_COMMAND_POSITION, &position) < 0)
    {
      ERROR("");
      perror("HCTL_IOC_SET_COMMAND_POSITION");
      goto end;
    }
  /* Actual position  0 */
  if (ioctl(fd, HCTL_IOC_RESET_ACTUAL_POSITION, NULL) < 0)
    {
      ERROR("");
      perror("HCTL_IOC_RESET_ACTUAL_POSITION");
      goto end;
    }
  /* Go to position mode */
  if (ioctl(fd, HCTL_IOC_POSITION_MODE, NULL) < 0)
    {
      ERROR("");
      perror("HCTL_IOC_POSITION_MODE");
      goto end;
    }
  /* Velocity = init */
  command_velocity_reg = pvelocity2reg((float)initial_velocity);
  if (ioctl(fd, HCTL_IOC_SET_PROPORTIONAL_VELOCITY, &command_velocity_reg) < 0)
    {
      perror("HCTL_IOC_SET_PROPORTIONAL_VELOCITY");
      goto end;
    }
  /* Go to proportional velocity mode */
  if (ioctl(fd, HCTL_IOC_START_PROPORTIONAL_VELOCITY, NULL) < 0)
    {
      ERROR("");
      perror("HCTL_IOC_START_PROPORTIONAL_VELOCITY");
      goto end;
    }

  /* Launch measure (blocking) */
  if (ioctl(fd, HCTL_IOC_PVELOCITY_MEASURE, &measure) < 0)
    {
      ERROR("");
      perror("HCTL_IOC_PVELOCITY_MEASURE");
      goto end;
    }

  /* Stop proportional velocity mode */
  if (ioctl(fd, HCTL_IOC_STOP_PROPORTIONAL_VELOCITY, NULL) < 0)
    {
      ERROR("");
      perror("HCTL_IOC_STOP_PROPORTIONAL_VELOCITY");
      goto end;
    }
  /* Go to idle mode */
  if (ioctl(fd, HCTL_IOC_IDLE_MODE, NULL) < 0)
    {
      ERROR("");
      perror("HCTL_IOC_IDLE_MODE");
      goto end;
    }
  
  /*
   * Output file
   */ 

  /* ouverture du fichier de commande */
  if(ofilename != NULL) {
    ostream = fopen(ofilename, "w");
    if (ostream == NULL) {
      ERROR("");
      perror(ofilename);
      goto end;
    }

    /* Ecriture des rsultats */
    for( index =0; index < measure.nb_measure; index++ )
      {
	fprintf(ostream, "%d ", measure.actual_velocity[index]);
      }
    fprintf(ostream, "\n");
    for( index =0; index < measure.nb_measure; index++ )
      {
	fprintf(ostream, "%lld ", (long long int)measure.position[index]);
      }
    fprintf(ostream, "\n");
  
    /* on ferme */
    fclose(ostream);
  }

  error = 0;

 end:
  free(measure.command_velocity);
  free(measure.command_duration);
  if( measure.measure_period > 0 )
    {
      free(measure.actual_velocity);
      free(measure.position);
    }
  return error;
}

/*
 * wizhctl-pvelocity entry
 */
int main(int argc, char **argv)
{
  struct arguments args;
  //wizhctl_pvelocity_t pvelocity;
  //FILE *ostream, *istream;
  int hctl_fd;
  int result;
  //int i;

  /* Default values. */
  args.verbose = 1;
  args.input_file = NULL;
  args.output_file = NULL;
  args.initial_velocity = 0;
  args.device = DEFAULT_DEVICE;
  args.period = 0;

  /* Parse our arguments; every option seen by parse_opt will
     be reflected in arguments. */
  argp_program_bug_address = bug_address;
  argp_program_version = program_version;
  argp_parse (&argp, argc, argv, 0, 0, &args);

  //printf("%s %s\n", args.input_file, args.output_file);

  /* Open hctl device */
  hctl_fd = open(args.device,  O_RDWR | O_SYNC);
  if (hctl_fd <= 0) {
    perror(args.device);
    exit(-1);
  }

  /* Le's go! */
  if( args.input_file == NULL )
    {
      /* If no input_file, launch interactive mode. */
      result = interactive_mode(hctl_fd, 
				(float)args.initial_velocity, 
				args.verbose);
    }
  else
    {
      /* If input_file, launch batch mode. */
      result = batch_mode(hctl_fd, 
			  args.input_file, 
			  args.output_file, 
			  args.period,
			  (float)args.initial_velocity, 
			  args.verbose);
    }

  close(hctl_fd);
  return result;
}
